@@ -14,7 +14,7 @@
 
 from __future__ import print_function
 
-if __name__=='pymol.editing':
+if True:
 
     import pymol
     import math
@@ -359,11 +359,6 @@ PYMOL API
 
         r = DEFAULT_ERROR
 
-        suspend_undo = _self.get("suspend_undo")
-        fin = 1
-        push_undo(target_name, just_coordinates=0, finish_undo=0, _self=_self)
-        _self.set("suspend_undo", 1, updates=0)
-
         if source_name == None:
             source_name = ''
 
@@ -389,14 +384,7 @@ PYMOL API
         finally:
             _self.unlock(r,_self)
         if _self._raising(r,_self):
-            _self.set("suspend_undo", suspend_undo, updates=0)
-            fin = -1
-            push_undo("", just_coordinates=0, finish_undo=fin, _self=_self)
             raise pymol.CmdException            
-
-            _self.set("suspend_undo", suspend_undo, updates=0)
-            push_undo("", just_coordinates=0, finish_undo=fin, _self=_self)
-
         return r
 
 
@@ -963,6 +951,8 @@ NOTES
     single molecular object.
 
 '''
+        import pymol.wizard.dragging
+
         quiet = int(quiet)
         if (selection!=None) and (selection!=""):
             selection = selector.process(selection)
@@ -991,7 +981,7 @@ NOTES
                 wiz = _self.get_wizard()
                 if (wiz == None):
                     _self.wizard("dragging",old_button_mode)
-                elif wiz.__class__ != 'pymol.wizard.dragging.Dragging':
+                elif not isinstance(wiz, pymol.wizard.dragging.Dragging):
                     _self.wizard("dragging",old_button_mode)
                 else:
                     wiz.recount()
@@ -1136,31 +1126,28 @@ DESCRIPTION
         return r
 
 
-    def h_add(selection="(all)",quiet=1,_self=cmd):
+    def h_add(selection="(all)", quiet=1, state=0, legacy=0, _self=cmd):
         '''
 DESCRIPTION
 
-    "h_add" uses a rudimentary algorithm to add hydrogens onto a
-    molecule based on current valences.
+    "h_add" adds hydrogens onto a molecule based on current valences.
 
 USAGE
 
-    h_add selection
+    h_add [ selection [, state ]]
 
 ARGUMENTS
 
     selection = string {default: (all)}
 
+    state = int {default: 0 (all states)}
+
 NOTES
 
     Because PDB files do not normally contain bond valences for
     ligands and other nonstandard components, it may be necessary to
     manually correct ligand conformations before adding hydrogens.
 
-PYMOL API
-
-    cmd.h_add( string selection="(all)" )
-
 SEE ALSO
 
     h_fill
@@ -1171,7 +1158,8 @@ SEE ALSO
         r = DEFAULT_ERROR
         try:
             _self.lock(_self)   
-            r = _cmd.h_add(_self._COb,selection,int(quiet))
+            r = _cmd.h_add(_self._COb,selection,int(quiet),
+                    int(state) - 1, int(legacy))
         finally:
             _self.unlock(r,_self)
         if _self._raising(r,_self): raise pymol.CmdException            
@@ -2047,8 +2035,8 @@ SEE ALSO
         target_name = str(target_name).strip()
         source_name = str(source_name).strip()
         if (target_name == '' and source_name != ''): # tentative -- create a new command instead?
-            mat = cmd.get_object_matrix(source_name,source_state)
-            view = cmd.get_view()
+            mat = _self.get_object_matrix(source_name,source_state)
+            view = _self.get_view()
             new_view = (
                 mat[ 0]*view[ 0] + mat[ 4]*view[ 3] + mat[ 8]*view[ 6],
                 mat[ 0]*view[ 1] + mat[ 4]*view[ 4] + mat[ 8]*view[ 7],  
@@ -2067,7 +2055,7 @@ SEE ALSO
                 mat[ 8]*view[12] + mat[ 9]*view[13] + mat[10]*view[14] -
                 mat[ 2]* mat[ 3] - mat[ 6]* mat[ 7] - mat[10]* mat[11],
                 view[15] ,         view[16] ,         view[17] )
-            r = cmd.set_view(new_view)
+            r = _self.set_view(new_view)
         else:
             try:
                 _self.lock(_self)
@@ -2539,26 +2527,30 @@ NOTES
 
     SPECIAL FLAGS
 
-    * Flags 0-7 are reserved for molecular modeling 
+    * Flags 0-5 are reserved for molecular modeling
 
         focus      0 = Atoms of Interest (i.e. a ligand in an active site) \\
         free       1 = Free Atoms (free to move subject to a force-field) \\
         restrain   2 = Restrained Atoms (typically harmonically contrained) \\
         fix        3 = Fixed Atoms (no movement allowed) \\
         exclude    4 = Atoms which should not be part of any simulation
+        study      5
+
+    * Flags 6-7 are for protein and nucleic acid classification
 
     * Flags 8-15 are free for end users to manipulate
 
-    * Flags 16-23 are reserved for external GUIs and linked applications
+    * Flags 16-21 are reserved for external GUIs and linked applications
 
-    * Flags 24-30 are reserved for PyMOL internal usage
+    * Flags 22-23 are for temporary use only (flag 23 used for coverage
+      tracking when assigning parameters in chempy.champ.assign)
+
+    * Flags 24-31 are reserved for PyMOL internal usage
 
         exfoliate 24 = Remove surface from atoms when surfacing \\
         ignore    25 = Ignore atoms altogether when surfacing \\
         no_smooth 26 = Do not smooth atom position
 
-    * Flag 31 is reserved for coverage tracking when assigning parameters, etc.
-
 PYMOL API
 
     cmd.flag(int flag, string selection, string action="reset",
@@ -2634,7 +2626,7 @@ SEE ALSO
             _self.disable(model)
 
         if group:
-            cmd.group(group, ' '.join(names_list), 'add')
+            _self.group(group, ' '.join(names_list), 'add')
 
     def alphatoall(selection='polymer', properties='b', operator='byca',
             quiet=1, _self=cmd):
@@ -2672,3 +2664,122 @@ DESCRIPTION
         if not int(quiet):
             print(' Altered %d MSE residues to MET' % (x))
 
+
+    def _base(i, numerals, _emptyzero=False):
+        if i == 0:
+            return '' if _emptyzero else numerals[0]
+
+        b = len(numerals)
+        return _base(i // b, numerals, True) + numerals[i % b]
+
+
+    def uniquify(identifier, selection, reference='', quiet=1, _self=cmd):
+        '''
+DESCRIPTION
+
+    Make `identifier` unique with respect to reference selection.
+
+ARGUMENTS
+
+    identifier = str: atom identifier (chain, segi, etc.)
+
+    selection = str: atom selection to modify
+
+    reference = str: atom selection whose identifiers must not be
+    present in the first selection {default: !selection}
+
+EXAMPLE
+
+    fetch 1a00 1hbb, async=0
+    uniquify chain, 1hbb
+    # 1hbb now has chains E,F,G,H
+        '''
+        import itertools
+        import string
+
+        if not reference:
+            reference = '!(' + selection + ')'
+
+        p = identifier
+
+        set_ref = set()
+        set_sel = set()
+        mapping = {}
+        space = {'set_ref': set_ref, 'set_sel': set_sel, 'mapping': mapping}
+
+        _self.iterate(reference, 'set_ref.add(' + p + ')', space=space)
+        _self.iterate(selection, 'set_sel.add(' + p + ')', space=space)
+
+        set_union = set_ref | set_sel
+        set_inter = set_ref & set_sel
+
+        if not set_inter:
+            return
+
+        baseargs = ()
+        i_iter = itertools.count(1)
+
+        if isinstance(next(iter(set_inter)), int):
+            basefunc = int
+        elif p in ('resi',):
+            basefunc = str
+        else:
+            basefunc = _base
+            baseargs = (string.ascii_uppercase + '123456789',)
+            i_iter = itertools.count(0)
+
+        for name in set_inter:
+            for i in i_iter:
+                newname = basefunc(i, *baseargs)
+                if newname not in set_union:
+                    mapping[name] = newname
+                    set_union.add(newname)
+                    break
+
+        _self.alter(selection, '%s = mapping.get(%s, %s)' % (p, p, p),
+                space=space)
+
+        if not int(quiet):
+            print(' Uniquify: renamed %d %s identifier(s)' % (len(mapping), p))
+
+
+    def copy_to(name, selection, rename='chain segi ID', zoom=-1, quiet=1, _self=cmd):
+        '''
+DESCRIPTION
+
+    Copies selection to object `name` (all states) and by default
+    renames chain, segi and ID identifiers to avoid naming conflicts.
+
+ARGUMENTS
+
+    name = str: object name to modify
+
+    selection = str: atom selection (will be copied to `name`)
+
+    rename = str: space separated list of identifiers to rename
+    {default: chain segi ID}
+
+SEE ALSO
+
+    create, fuse
+        '''
+        temp = _self.get_unused_name('_tmp')
+
+        try:
+            _self.create(temp, selection, zoom=0)
+            _self.disable(' '.join(_self.get_object_list(selection)))
+
+            for prop in rename.split():
+                if prop.upper() == 'ID':
+                    _self.alter(temp, 'ID = -1')
+                else:
+                    uniquify(prop, temp, '?' + name, quiet=quiet)
+
+            _self.create(name, '?' + name + ' ' + temp, zoom=zoom)
+            _self.unpick()
+
+            if not int(quiet):
+                n = _self.count_atoms(temp)
+                print(' Copied %d atoms to object %s' % (n, name))
+        finally:
+            _self.delete(temp)
